/*

    sftpd - meant to be a small and flexible ftp server implementation
    Copyright (C) 2009  Jan Murawski  der@oelgoetze.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

    For more information about the GNU Public license see:
    http://www.gnu.org/copyleft/gpl.html

*/

#ifndef _DIRECTORY_OBJ_
#define _DIRECTORY_OBJ_


#include <deque>
#include <vector>
#include <string>
using namespace std;
#include <windows.h>
#include "fileUtil.h"
#include "timeUtil.h"
#include "convertUtil.h"


//
// struct, to enabled seperation of information
// capturing and sorting by filename
//
struct searchEntry {
    string fileSize;
    string timeStamp;
    string fileName;
    string lowerCaseFileName;   // convert string to lower chars for comparisson
    string dirMask;
    bool isDirectory;
    bool isReadOnly;
    bool isExecutable;
};


//
// current server path consists of one root node 
// which may be followed by others
// (new nodes can be created to handle .lnk for example)
//
struct dirNode {
    string rootDir;
    string virtuaRootDir;
    dirNode *previousNode;
    dirNode *nextNode;
};


//
// object which maps virtual server path to real local path
//
class directoryObj
{
    bool writable;          // user has write access
    dirNode *rootNode;      
    string currVirtualDir;  // current virtual dir user is in
    string currDir;         // current real dir     
    
    void removeSubnodes();  // removes all dir nodes except for head node
    bool goDirUp();         // steps a dir up. removes last dir-node, if no more needed
                            // returns false, whend absolut root is reached
    bool goDirDown(string folderName);

    bool removeLastDirFromPath(string &path,            // if any more existant, remove last folder from path
                               bool unixStyle);         // designed form absout paths only
                                                        // returns true, when dir removed   
    void finalBackSlashToPath();                        // append final backslash to path if non existant
    void finalSlashToPath();                            // append final slash to path if non existant
    void finalSlashToPath(string &path);


public:
    directoryObj(string directory, string writable);
    ~directoryObj();

    bool getPwd(string &pwd);   // returns pwd
                                // pwd is changend upwards until a folder is reached
                                // which is existant

    bool getDirListing(string userName, vector<string*> *lines);    // returns a complete directory listing
    bool changeDir(string vDir);                // changes to a given relative or absolut path
    bool fileNameToFullPath(string &fileName);  // converts a given filename in current folder to its full path
    bool getRealPath(string &virtualPath);      // converts a virtual to a real path
};




#endif _DIRECTORY_OBJ_
