/*

    sftpd - meant to be a small and flexible ftp server implementation
    Copyright (C) 2009  Jan Murawski  der@oelgoetze.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

    For more information about the GNU Public license see:
    http://www.gnu.org/copyleft/gpl.html

*/

#include "fileUtil.h"


//
// Test if a file exists
//
bool fileExists(string file)
{
    char *buffer = new char[file.length() +1];
    for (unsigned int i = 0; i < file.length(); i++)
        *(buffer + i) = file[i];
    *(buffer + file.length()) = 0;

    bool retVal;
    if (PathFileExists(buffer))
        retVal = true;
    else
        retVal = false;

    delete buffer;
    return retVal;
}


//
// Check, if given file is executable (.exe, .bat, etc.)
//
bool IsExecutable(char *path)
{
    SHFILEINFO *fi = new SHFILEINFO();
    int flags = SHGFI_EXETYPE;     // return exe type

    // FIXME: function call is quite slow !?
    int res = SHGetFileInfo(
        path,
        0,
        fi,
        sizeof(SHFILEINFO),
        SHGFI_EXETYPE);

    delete fi;
    return (res != 0);
}

bool IsExecutable(string path)
{
    char *buffer = new char[path.size() +1];
    for (unsigned int i = 0; i < path.size(); i++)
        *(buffer +i) = path.at(i);
    *(buffer + path.size()) = 0;

    bool retVal = IsExecutable(buffer);

    delete buffer;
    return retVal;
}


//
// calculate a files size in 64 bit
//
long long getFileSize(string fileName)
{
    // copy the filename to a buffer
    char *filename = new char[fileName.size() +1];
    for (unsigned int i = 0; i < fileName.size(); i++)
        *(filename + i) = fileName.at(i);
    *(filename + fileName.size()) = 0;

    // vars for searching
    WIN32_FIND_DATA FindFileData;
    HANDLE hFind = INVALID_HANDLE_VALUE;

    // Find the first file in the directory.
    hFind = FindFirstFile(filename, &FindFileData);

    if (hFind == INVALID_HANDLE_VALUE) 
    {
        //printf ("Invalid file handle. Error is %u.\n", GetLastError());
        // FIXME: error logging
        delete filename;
        return -1;
    } 

    // reasemble 64 bit filesize
    long long fileSize;
    int32To64(FindFileData.nFileSizeLow, FindFileData.nFileSizeHigh, fileSize);

    FindClose(hFind);   // search is finished

    // clean up and return
    delete filename;
    return fileSize;
}

string getFileSizeAsString(string fileName)
{
    return intToStr(getFileSize(fileName));
}
