/*

    sftpd - meant to be a small and flexible ftp server implementation
    Copyright (C) 2009  Jan Murawski  der@oelgoetze.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

    For more information about the GNU Public license see:
    http://www.gnu.org/copyleft/gpl.html

*/

#include "serverObjList.h"


// export functions, if build as dll
#undef  WS2_32_API
#ifdef _EXPORT_DLL_  
    #define WS2_32_API __declspec(dllexport)
#else
    #define WS2_32_API
#endif _EXPORT_DLL_ 


CRITICAL_SECTION csServerObjInAccess;   // sync acc to obj list
serverListObj* firstServerObj;          // first element in our (thread-) gloabl obj list
bool initialized = false;


//
// init our (thread-) gloabl list
// MUST BE CALLED FIRST  -  app will crash elsewise
//
void initServerObjList()
{   
    if (!initialized)
    {
        initialized = true;
        firstServerObj = 0;     // no objects there
        InitializeCriticalSection(&csServerObjInAccess);
    }
}


//
// add a new server obj to our (thread-) gloabl list
// returns obj address as a handle
//
WS2_32_API int newServerObj()
{
    EnterCriticalSection(&csServerObjInAccess);

    // init our new obj
    serverListObj* newObj = new serverListObj;
    newObj->next = 0;
    
    if (firstServerObj == 0)
        firstServerObj = newObj;    // obj now only obj in list
    else
    {
        serverListObj* currObj = firstServerObj;
        while (currObj->next != 0)
            currObj = currObj->next;

        currObj->next = newObj;     // obj now last in list
    }
    
    LeaveCriticalSection(&csServerObjInAccess);
    return (int) newObj;
}


//
// remove a server obj from our (thread-) gloabl list
// will faild if addr is not valid
// will return as soon as server has exited
//
WS2_32_API bool removeServerObj(int objAddr)
{
    serverListObj *oldObj = (serverListObj*) objAddr;   // obj to remove
    serverListObj *lastObj, *currObj;

    if (!isValidServerObj(objAddr)) // objAddr valid ?
            return false;

    EnterCriticalSection(&csServerObjInAccess); // isValidServerObj will deadlock if called before !!

    if (oldObj == firstServerObj)
        firstServerObj = 0;     // no more objects left
    else
    {
        // lets search the obj to remove
        lastObj = firstServerObj;
        currObj = lastObj->next;
        while (currObj != oldObj)
        {
            lastObj = currObj;
            currObj = currObj->next;
            
        }
        if (currObj->next != 0)
            lastObj->next = currObj->next;  // previous obj now links to obj following uns
        else
            lastObj->next = 0;  // previous obj is now last obj in list

    }

    // clean up
    delete(oldObj);

    LeaveCriticalSection(&csServerObjInAccess);
    return true;
}


//
// returns true, if given addr is a valid handle
//
WS2_32_API bool isValidServerObj(int objAddr)
{
    EnterCriticalSection(&csServerObjInAccess);

    if (firstServerObj == 0)    // list is empty => addr must be wrong
    {
        LeaveCriticalSection(&csServerObjInAccess);
        return false;
    }


    // search list for given obj
    serverListObj *currObj = firstServerObj;
    do
    {
        if (currObj == (serverListObj*) objAddr)
        {
            LeaveCriticalSection(&csServerObjInAccess);
            return true;
        }
        currObj = currObj->next;
    }
    while (currObj->next != 0);


    LeaveCriticalSection(&csServerObjInAccess);
    return false;
}
