/*

    sftpd - meant to be a small and flexible ftp server implementation
    Copyright (C) 2009  Jan Murawski  der@oelgoetze.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

    For more information about the GNU Public license see:
    http://www.gnu.org/copyleft/gpl.html

*/

#include "serverThreads.h"


//
// hanlde incoming connections. 
// starts one new thread per controle connection and
// registers data connections with matching handles
//
DWORD WINAPI listenThread(LPVOID args)
{
    // start listenning on controle and data port
    listen(*(*(listenThreadInfo*)args).ctrlListenSock, SOMAXCONN);
    listen(*(*(listenThreadInfo*)args).dataListenSock, SOMAXCONN);

    // If iMode != 0, non-blocking mode is enabled.
    u_long iMode = 1;
    ioctlsocket(*(*(listenThreadInfo*)args).ctrlListenSock, FIONBIO, &iMode);
    ioctlsocket(*(*(listenThreadInfo*)args).dataListenSock, FIONBIO, &iMode);

    SOCKET *TempSock = new SOCKET;
    bool connectionMade;
    do
    {
        connectionMade = false;

        // new incoming CONTROLE connections pending?
        *TempSock = accept(*(*(listenThreadInfo*)args).ctrlListenSock,NULL,NULL);
        if (*TempSock != SOCKET_ERROR)
        {
            // prepare struct to parse information to thread
            threadInfo *ti = new threadInfo;
            ti->svObj = (*(listenThreadInfo*)args).svObj;
            ti->sck = TempSock;

            // start new session thread
            {
                DWORD threadId;
                HANDLE h = CreateThread(NULL, 0,  &sessionThread, ti, 0, &threadId);
                (*(*(listenThreadInfo*)args).svObj).regThread(h);
            }

            // old one in use - get us a new sck obj
            TempSock = new SOCKET;

            connectionMade = true;
        }

        // ensure, that we want some new connection, before we accept one.
        // this is a work around for pushing clients, who connect imediatly after
        // the pasv command and don't wait for our "150 socket is ready", during
        // the list command, like proposen by RFC #959
        //                              FIXME:
        // a full tcp handshake is made by windoze even if we do not accept at once.
        // (something about syn flood protection)
        // i really hope windows doesn't mind this. haven't run into problems yet :)
        // but all in all this may lead to tcp connections staying open till a regular data
        // connection is expected to arrive. potential for a dos attack?
        if ((*(*(listenThreadInfo*)args).svObj).getDataConnHandleCount() > 0)
        {
            // new incoming DATA connections pending?
            *TempSock = accept(*(*(listenThreadInfo*)args).dataListenSock, NULL, NULL);
            if (*TempSock != SOCKET_ERROR)
            {
                char *remoteIp = 0;
                getRemoteSockIp(TempSock, &remoteIp);
                
                dataConnHandle *dch = (*(*(listenThreadInfo*)args).svObj).findDataConnHandle(remoteIp);

                // if connection is not welcome (we are not aware that a client might connect)
                // or another socket has already been registered with handle, close socket and free mem
                if (dch == 0
                    || !(*(*(listenThreadInfo*)args).svObj).registerSocketWithDCH(dch, TempSock))
                {
                    closesocket(*TempSock);
                    delete TempSock;
                }
                // old one in use or removed, so get us a new sck obj
                TempSock = new SOCKET;

                connectionMade = true;
            }
        }

        // if necessary, mark data connection reqeusts as timedout
        (*(*(listenThreadInfo*)args).svObj).maintainDataConnHandleList();

        if (!connectionMade)
            // nothing new *zzZZzzZz*
            // save some cpu time
            Sleep(1);
    }
    // resume till daddy wants us to stop
    while((*(*(listenThreadInfo*)args).svObj).getGetFinished() == false);

    delete(args);

    ExitThread(0);
    return 0;
}


//
//  searches for the next command in given buffer
//  command always end with \n || \r\n
//
void getNextCommand(string *buffer, string *command)
{
    int pos = buffer->find("\n");

    if (pos  == string::npos)
        *command = "";
    else
    {
        *command = buffer->substr(0, pos -1 +1);
        buffer->erase(0, pos +1);

        // if client sends windows newline \r\n, then \n is already removed
        // remove \r now
        // (\r\n is rfc conform :S)
        if (command->length() > 0)
            if (command->substr(command->length() -1, 1) == "\r")
                command->erase(command->length() -1);
    }
}


//
// Does strIs begin whith strShouldBe ?
// If yes, remove beginning and return true.
//
bool isCommand(string *strIs, string *strShouldBe)
{
    if (strIs->substr(0, strShouldBe->length()).compare(*strShouldBe) == 0)
    {
        strIs->erase(0, strShouldBe->length());
        return true;
    }

    return false;
}

#define ISCOMMAND(is, shouldBe)  \
        isCommand(is, &string(shouldBe))



//
// requests a data connection handle and waits, till it gets active
//
bool getActiveHandle(LPVOID args, SOCKET *controleSock, dataConnHandle **dch)
{
    // get client ip
    char *remoteIp = 0;
    if (!getRemoteSockIp(controleSock, &remoteIp))
        return false;
    
    // tell listen thread to wait for connection
    *dch = (*(*(listenThreadInfo*)args).svObj).createDataConnHandle(remoteIp);
    
    // if there is another thread waiting for a dataconnection from the same ip
    // wait till it has been removed from queue == we are first in search
    // (notice remote host every 10 seconds about pending connection  <= removed)
    int times = 0;
    while ((*(*(listenThreadInfo*)args).svObj).findDataConnHandle(remoteIp) != *dch)
    {
        Sleep(1000);    // time to sleep should be smaller than connection timeout
                        // to ensure timer is resetted in time

        // note: we do not message the client, cause dump clients will falsely rcognise this as
        // our 150 or some later status code.
        // this is NOT compliant to RFC #959

        // reset timer to ensure, we play fair ^^
        (*(*(listenThreadInfo*)args).svObj).resetDataConnHandleTimer(*dch);
    }

    return true;
}


//
// wait till a expected data connection arives,
// global finish is requested or the timeout is reached
//
int waitForDataConnArrival(LPVOID args, dataConnHandle *dch)
{
    bool registered = false;
    bool timedOut = false;
    bool globalGetFinished = false;

    // wait for connection request to time out or to be made
    while ( !(timedOut = (*(*(listenThreadInfo*)args).svObj).isDataConnHandleTimedOut(dch))
        &&  !(registered = (*(*(listenThreadInfo*)args).svObj).isSocketRegistered(dch))
        &&  !(globalGetFinished = (*(*(listenThreadInfo*)args).svObj).getGetFinished()))
    {
        Sleep(1);
    }

    // remove from queue
    (*(*(listenThreadInfo*)args).svObj).deleteDataConnHandleEntry(dch);

    // convert result for caller
    if (globalGetFinished)
        return WAITFORDATACONN_ERR_GLOBAL_FINISH;
    else if (timedOut)
        return WAITFORDATACONN_ERR_TIMEOUT;
    else
        return WAITFORDATACONN_ERR_NOERR;
}


//
// functional core of the server
//
DWORD WINAPI sessionThread(LPVOID args)
{
    threadInfo *ti = (threadInfo*)args;     // new var to simplify pointer args handling
    SOCKET *controleSock = ti->sck;
    int retVal = SESSION_THREAD_ERR_NOERR;

    // If iMode != 0, non-blocking mode is enabled.
    u_long iMode = 1;
    ioctlsocket(*controleSock, FIONBIO, &iMode);

    string inBuf = "";              // string to manage incomming commands (data)
    string command = "";            // string to save current command in
    string commandBackup = "";      // backup copy of current command - to become lastCommand
    string lastCommand = "";        // remember the last command handled (for REST e.g.)

    bool loggedIn = false;
    string user = "";
    string pw = "";

    directoryObj *dObj = 0;             // object to handle directory structure

    int lastActionAt = GetTickCount();  // remember time of last contact - and kick if passive for to long
    int currentActionAt;
    int lastActionLastet = 0;           // calculate time used for handling last command 
                                        // so that it does not count as idele time
    bool commandHandlingDone = false;   // shows, if a command handling has been done
                                        // at the end of last loop

    char type = 'I';                // transfer data representation type
    char stru = 'F';                // file structure for transfer
    char mode = 'S';                // transfer mode

    long long restMarker = 0;       // positon at which RETR and STOR start
    bool pasvModeSet = false;       // already set to pasive mode?

    bool cmdLIST, cmdRETR;          // let's remember what action is going on ^^

    HANDLE file = INVALID_HANDLE_VALUE; // handle for file actions. will be closed at the beginning of main loop 

    // send first hello
    if (!SENDTOSOCKETLN(controleSock, "220 Service ready for new user."))
    {
        // FIXME: error logging
        retVal = SESSION_THREAD_ERR_WRITE;
        goto end;
    }

    // main session loop from here:
    do
    {

        // calculate the time which the command handling took
        // so that it does not count as idel time
        if (commandHandlingDone)
        {
            lastActionLastet = GetTickCount() - lastActionAt;
            commandHandlingDone = false;
        }

        // return file handle if still opened
        if (file != INVALID_HANDLE_VALUE)
        {
            CloseHandle(file);
            file = INVALID_HANDLE_VALUE;
        }

        if (readFromSocket(controleSock, &inBuf) != true)
        {
            // FIXME: error logging
            retVal = SESSION_THREAD_ERR_READ;
            break;
        }
    
        getNextCommand(&inBuf, &command);

        // no command ready for us?
        if (command == "")
        {
            // does next command exceed max len
            if (inBuf.length() >= MAX_BUF_LEN)
            {
                SENDTOSOCKETLN(controleSock, "500 Command to long.");
                inBuf = "";
            }
            else
            {
                // ensure the time which the last command handling took
                // does not count as idle time
                if (lastActionLastet > 0)
                {
                    lastActionAt += lastActionLastet;
                    lastActionLastet = 0;
                }

                // terminate connection, if idle for to long..
                currentActionAt = GetTickCount();
                if (currentActionAt > lastActionAt + WAIT_FOR_NEXT_CMD_TIMEOUT)
                {
                    retVal = SESSION_THREAD_ERR_TIMEOUT;
                    break;
                }

                Sleep(1);
            }
            continue;
        }
        else
            // reset timer
            lastActionAt = GetTickCount();


        // convert the first part of the command up to the first space to
        // upper case chars, so we can handle non standard conform clients
        // like the m$ vista one for example, too.
        int pos;
        if ((pos =  command.find(' ')) != string::npos)
        {
            pos--;
            for (unsigned int i = 0; (int) i < pos; i++)
                command[i] = toupper(command[i]);
        }

        // store a copy of current command to make it lastCommand for next loop
        commandBackup = command;
        
        // remember, that we handled a command to substract time consumed for it
        // at the beginning of next loop
        // this enables command handling to do continues
        commandHandlingDone = true;

// ******************************                       ******************************
// ******************************   user NOT logged in  ******************************
// ******************************                       ******************************

        if (!loggedIn)
        {

// ****************************** set user ******************************

            if (ISCOMMAND(&command, "USER "))
            {
                user = command;
                SENDTOSOCKETLN(controleSock, "331 Password required.");
            }

// ****************************** set password ******************************

            else if (ISCOMMAND(&command, "PASS ") || ISCOMMAND(&command, "PASS"))   // ISCOMMAND important here for 
                                                                                    // logins whith blank passwords
            {
                pw = command;
                if (ti->svObj->loginValid(user, pw))
                {
                    // initialise virtual directory structure
                    string directory, rw;
                    ti->svObj->getLoginSettings(user, pw, directory, rw);
                    dObj = new directoryObj(directory, rw);

                    loggedIn = true;
                    SENDTOSOCKETLN(controleSock, "230 User logged in, proceed.");
                }
                else
                {
                    SENDTOSOCKETLN(controleSock, "530 Login incorrect.");
                }
            }

// ****************************** log out and disconnect ******************************

            else if (command == "QUIT")
            {
                SENDTOSOCKETLN(controleSock, "221 Goodbye.");
                break;
            }
            
// ****************************** no other command allowed until login ******************************

            else
                SENDTOSOCKETLN(controleSock, "530 Not logged in.");
        }

// ******************************                   ******************************
// ******************************   user logged in  ******************************
// ******************************                   ******************************

        else
        {

// ****************************** print working directory ******************************

            if (command == "PWD" || command == "XPWD")
            {
                string pwd;
                dObj->getPwd(pwd);
                sendToSocketLn(controleSock, "257 \"" + pwd + "\" is current directory.");
            }

// ****************************** change directory ******************************
            
            else if (ISCOMMAND(&command, "CWD "))
            {
                if (command == "" || !dObj->changeDir(command))
                    SENDTOSOCKETLN(controleSock, "550 Failed to change directory.");
                else
                    SENDTOSOCKETLN(controleSock, "250 Directory successfully changed.");
            }

// ****************************** change to parent directory ******************************
            
            else if (command == "CDUP")
            {
                if (!dObj->changeDir("../"))
                    SENDTOSOCKETLN(controleSock, "550 Failed to change directory.");
                else
                    SENDTOSOCKETLN(controleSock, "250 Directory successfully changed.");
            }

// ****************************** get directory listing ******************************
// ****************************** retrieve a file ******************************


            else if ((cmdLIST = (command == "LIST")) | 
                     (cmdRETR = (ISCOMMAND(&command, "RETR "))))
            {
                if (!pasvModeSet)
                {
                    // "Illegal PORT command."
                    SENDTOSOCKETLN(controleSock, "425 Use PASV first.");    // Use PORT or PASV first.
                    continue;
                }

                // dump clients wont recognize informational codes -.-
                // so we do not send an information, if the type is set to
                // something else than I (and assume it is meant to be I)

                string fileName = "";
                if (cmdRETR)
                {
                    fileName = command;

                    // get us the full pathname and check if the file exists and open it
                    if (!dObj->getRealPath(fileName) ||
                        !fileExists(fileName) ||
                        (file = CreateFile(fileName.c_str(),
                                           GENERIC_READ,                // access mode
                                           FILE_SHARE_READ |            // actions allowed for other processes
                                           FILE_SHARE_WRITE |
                                           FILE_SHARE_DELETE,
                                           0,                            // cannot be inherited by any child processes
                                           OPEN_EXISTING,               // open mode
                                           FILE_FLAG_SEQUENTIAL_SCAN,   // flags for new files
                                           0)                           // handle to a template file
                        ) == INVALID_HANDLE_VALUE)
                    {
                        SENDTOSOCKETLN(controleSock, "550 Failed to open file.");
                        continue;
                    }

                    // if we didn't follow immediatly after the REST command
                    // ignore value set by REST to be RFC conform
                    if (!ISCOMMAND(&lastCommand, "REST"))
                        restMarker = 0;
                }

                // reqeust a new handle from listen thead
                dataConnHandle *dch;
                if (!getActiveHandle(args, controleSock, &dch))
                {
                    retVal = SESSION_THREAD_ERR_CL_ADDR;
                    break;
                }

                // tell remote host we are good to go
                if (cmdLIST)
                    SENDTOSOCKETLN(controleSock, "150 Opening ASCII mode data connection for listing.");
                else if (cmdRETR)
                    // ready for   fielename  (filesize)
                    sendToSocketLn(controleSock, "150 Opening BINARY mode data connection for "
                      + command + " ("  + getFileSizeAsString(fileName) + ").");

                // wait for client to connect
                int result = waitForDataConnArrival(args, dch);

                if (    result == WAITFORDATACONN_ERR_TIMEOUT 
                     || result == WAITFORDATACONN_ERR_GLOBAL_FINISH )
                {
                    // connection failed, so clean up
                    (*(*(listenThreadInfo*)args).svObj).cleanUpDataConnHandle(dch);

                    // if daddy wants us to quit, leave main loop
                    if (result == WAITFORDATACONN_ERR_GLOBAL_FINISH)
                        break;

                    // inform client about timed out connection
                    SENDTOSOCKETLN(controleSock, "425 Unable to open data connection.");
                }
                else    // registered == true
                {
                    // we have got our socket
                    // let's handle transfer ...

                    // If iMode != 0, non-blocking mode is enabled.
                    u_long iMode = 1;
                    ioctlsocket(*(dch->sck), FIONBIO, &iMode);

//                 *********** directory listing ***********

                    if (cmdLIST)
                    {
                        // get file listing
                        vector<string*> *list = new vector<string*>;
                        bool errorRaised = false;
                        if (dObj->getDirListing(user, list))
                        {
                            for (unsigned int i = 0; i < list->size(); i++)
                            {   
                                // are we allowed to continue?
                                if ((*(*(listenThreadInfo*)args).svObj).getGetFinished())
                                {
                                    errorRaised = true;
                                    break;
                                }

                                // get line from listing and write to socket
                                string *currBuff = list->at(i);
                                if (!sendToSocket(dch->sck, *currBuff))
                                {
                                    errorRaised = true;
                                    SENDTOSOCKETLN(controleSock, "426 Failure writing network stream.");
                                    break;
                                }

                                // delete transfered buffer
                                delete currBuff;
                            }
                        }
                        else
                        {
                            errorRaised = true;
                            SENDTOSOCKETLN(controleSock, "426 Not a folder or not accessible.");
                        }

                        delete list;

                        if (!errorRaised)
                            SENDTOSOCKETLN(controleSock, "226 Transfer complete.");
                    }

//                 *********** retrieving a file ***********

                    else if (cmdRETR)
                    {
                        bool errorRaised = false;
                        
                        // file should already be open here
                        
                        // get file size
                        long long bytesToRead;
                        {
                            PLARGE_INTEGER lpFileSize = new LARGE_INTEGER;
                            GetFileSizeEx(file, lpFileSize);
                            bytesToRead = lpFileSize->QuadPart;
                        }

                        // keep restart marker in mind
                        long long bytesRead = restMarker;

                        // seek restart marker in file. igonre potential errors
                        // cause they'll be handled in the following loop either way
                        {
                            long long position = 0;

                            while (position != restMarker)
                            {
                                // distance to still to seek
                                long long difference = restMarker - position;
                                // needs to fit into 32 bit cause of api limitation
                                if (difference > MAXDWORD)
                                    difference = MAXDWORD;

                                // remove upper 32 bit
                                // FIXME : casting possible ?
                                long low, high;
                                int64To32(difference, low, high);

                                // seek 
                                DWORD move = low;
                                DWORD dummy = 0;
                                ReadFile(file, NULL, move, &dummy, NULL);

                                // update current position
                                position += difference;
                            }
                        }

                        rBufferHelper rBuffer(1024 * 1024 * 2);     // 2MB buffer

                        while (!(*(*(listenThreadInfo*)args).svObj).getGetFinished() &&
                                (bytesToRead > bytesRead || rBuffer.usedSpace() > 0))
                        {
                            // how many bytes will we read?
                            unsigned int count = 1024 * 512;  // read 512KB snippets

                            if (bytesToRead > 0)
                            { 
                                // enough bytes left?
                                if ((long long) count > bytesToRead - bytesRead)
                                    count = (int) (bytesToRead - bytesRead);
                                // enough space in buffer?
                                if (count > rBuffer.maxWritable())
                                    count = rBuffer.maxWritable();

                                // read n bytes from file
                                DWORD countRead = 0;
                                if (!ReadFile(file, 
                                              rBuffer.getEmptyBuffer(),
                                              count,
                                              &countRead,
                                              NULL))
                                {
                                    // FIXME: error logging
                                    errorRaised = true;
                                    SENDTOSOCKETLN(controleSock, "426 Failed to read from file.");
                                    break;
                                }

                                count = countRead;
                                rBuffer.added(count);
                                bytesRead += count;
                            }

                            int bSend = 0;
                            if (rBuffer.maxReadable() > 0)
                            {
                                // parse maximum buffer possible to winsock
                                bSend = send(*(dch->sck), 
                                    rBuffer.getUsedBuffer(), 
                                    rBuffer.maxReadable(), 
                                    0); 

                                if (bSend == SOCKET_ERROR)
                                {
                                    unsigned int errorNo;
                                    if ((errorNo = WSAGetLastError()) == WSAEWOULDBLOCK)
                                        // not enough buffer space available
                                        // so simply wait
                                        bSend = 0;
                                    else
                                    {
                                        // FIXME: error logging
                                        errorRaised = true;
                                        SENDTOSOCKETLN(controleSock, "426 Failure writing network stream.");
                                        break;
                                    }
                                }

                                // skip bytes taken from buffer
                                rBuffer.removed(bSend);
                            }

                            // take a break to let the other threads work, if theres nothting left to do
                            if (count == 0 && bSend == 0)
                                Sleep(1);

                            // note: this is not compliant to RFC #959.
                            // should be listenning for stauts commands via controle
                            // connection here
                        }

                        if (!errorRaised)
                            SENDTOSOCKETLN(controleSock, "226 File send OK.");
                    }

//                 **********************

                    // clean up
                    closesocket(*(dch->sck));
                    (*(*(listenThreadInfo*)args).svObj).cleanUpDataConnHandle(dch);
                }
            }

// ****************************** change the resart marker ******************************

            else if (ISCOMMAND(&command, "REST ") || command == "REST")
            {
                // read new restart marker
                // when told rubish use 0 as default
                if (command.size() == 0
                    || (restMarker = _atoi64(command.c_str())) < 0)
                {
                    restMarker = 0;
                }
                sendToSocketLn(controleSock, "350 Restart position accepted ("
                      + intToStr(restMarker) + ").");
            }

// ****************************** set data representation type for data connection ******************************

            else if (ISCOMMAND(&command, "TYPE "))
            {
                if (command == "I")
                {
                    type = 'I';
                    SENDTOSOCKETLN(controleSock, "200 Type set to I.");
                }
                else if (command == "A")
                {
                    type = 'A';
                    SENDTOSOCKETLN(controleSock, "200 Type set to A.");
                }
                else
                    SENDTOSOCKETLN(controleSock, "504 Command not implemented for that parameter.");
            }

// ****************************** set file structure type for data connection ******************************

            else if (ISCOMMAND(&command, "STRU "))
            {
                if (command == "F")
                {
                    stru = 'F';
                    SENDTOSOCKETLN(controleSock, "200 Structure set to F.");
                }
                else
                    SENDTOSOCKETLN(controleSock, "504 Command not implemented for that parameter.");
            }

// ****************************** set transfer mode for data connection ******************************

            else if (ISCOMMAND(&command, "MODE "))
            {
                if (command == "S")
                {
                    mode = 'S';
                    SENDTOSOCKETLN(controleSock, "200 Mode set to S.");
                }
                else
                    SENDTOSOCKETLN(controleSock, "504 Command not implemented for that parameter.");
            }

// ****************************** set server to passive mode ******************************

            // reply with ip and port to connect to
            else if (command == "PASV")
            {
                // retrive local ip
                string localIp;
                if (!getLocalSockIp(ti->sck, localIp))
                {
                    retVal = SESSION_THREAD_ERR_CL_ADDR;
                    break;
                }
                
                // format ip:
                // (a1,a2,a3,a4,p1,p2)
                // where a1.a2.a3.a4 is the IP address and p1*256+p2 is the port number. 
                int pos;
                while ((pos = localIp.find('.')) >= 0)
                    localIp[pos] = ',';
                
                // format date port nummber
                int p2 = ti->svObj->getDataPort() % 256;
                int p1 = (ti->svObj->getDataPort() - p2) / 256;
                // and convert to string
                string strP1 = intToString(p1);
                string strP2 = intToString(p2);

                sendToSocketLn(controleSock, "227 Entering Passive Mode (" + localIp + "," + strP1 + "," + strP2 + ")");
                pasvModeSet = true;
            }

// ****************************** report system type ******************************

            else if (command == "SYST")
            {
                // file listing is in unix style
                // on char is 8 bit long
                SENDTOSOCKETLN(controleSock, "215 UNIX Type: L8");
            }

// ****************************** no operation ******************************

            else if (command == "NOOP")
            {
                SENDTOSOCKETLN(controleSock, "200 NOOP ok.");
                break;
            }

// ****************************** log out and disconnect ******************************

            else if (command == "QUIT")
            {
                SENDTOSOCKETLN(controleSock, "221 Goodbye.");
                break;
            }

// ****************************** unknown command ******************************

            else
            {
                SENDTOSOCKETLN(controleSock,"502 Command not implemented or"
                                            " not implemented for that argument.");
            } // end of command selection
        } // end of logged in vs not logged in selection

        // remember current command (e.g. for RETR)
        lastCommand = commandBackup;
    }
    // resume main loop till daddy wants us to stop
    while((*(*(listenThreadInfo*)args).svObj).getGetFinished() == false);

end:
    closesocket(*(controleSock));
    delete controleSock;

    if (dObj != 0) delete dObj;
    delete(ti);

    return retVal;
}
