/*

    sftpd - meant to be a small and flexible ftp server implementation
    Copyright (C) 2009  Jan Murawski  der@oelgoetze.de

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

    For more information about the GNU Public license see:
    http://www.gnu.org/copyleft/gpl.html

*/

#include "socketHelper.h"


//
// read from winsock while there are bytes to read 
// and buffer does not exceed given size
//
bool readFromSocket(SOCKET *sck, string *buffer)
{
    char recvbuf[512];                          // buffer for reading from wsock
    int recvbuflen =                            // buffer size
        sizeof(recvbuf) / sizeof(recvbuf[0]);
    int bRead;                                  // bytes read

    bRead = 1;
    while (bRead > 0 && buffer->length() < MAX_BUF_LEN)
    {
        // get bytes from wsock
        bRead = recv(*sck, recvbuf, recvbuflen, 0);

        int error = WSAGetLastError();
        if (error != WSAEWOULDBLOCK && bRead == 0)
            // FIXME: error logging
            return false;

        // append bytes to buffer string
        for (int i = 0; i < bRead; i++)
            *buffer += *(recvbuf + i);
    }

    return true;
}


//
// send buffer to winsock
//
bool sendToSocket(SOCKET *sck, char *buffer, int len)
{
    int bSend = 0;
    unsigned int bytesSendTotal = 0;
    do
    {
        // we need to wait for winsock, if this 
        // isn't our first try to send, so wait
        if (bytesSendTotal != 0)
            Sleep(1);

        bSend = send(*sck, 
            buffer + bytesSendTotal,    // skip parts, that already have been send 
            len -1 -bytesSendTotal,     // strip string terminatin NULL 
            0); 

        if (bSend == SOCKET_ERROR)
        {
            unsigned int errorNo;
            if ((errorNo = WSAGetLastError()) == WSAEWOULDBLOCK)
                // not enough buffer space available
                bSend = 0;
            else
            {
                // FIXME: error logging
                //assert(false);
                return false;
            }
        }
        bytesSendTotal += bSend;

        if (bSend == 0)
            Sleep(1);
    }
    // resume, till complete buffer has been 
    // transfered to winsock (wich is != has been send)
    while (bytesSendTotal != len -1); 

    return true;
}

bool sendToSocket(SOCKET *sck, string buffer)
{
    char *buf = new char[buffer.length() +1];

    unsigned int i;
    for (i = 0; i < buffer.length(); i++)
        *(buf + i) = buffer[i];
    *(buf + i) = 0;

    bool retVal = sendToSocket(sck, buf, i+1);
    delete buf;
    return retVal;
}

//   use define SENDTOSOCKETLN(sck, data) instead
/*  
bool sendToSocketLn(SOCKET *sck, char *buffer, int len)
{
    if      (sendToSocket(sck, buffer, len -1)  // do net send NULL string terminator :)
        &&   sendToSocket(sck, "\r\n", 2))
    {
        return true;
    }
    else
        return false;

}
*/

bool sendToSocketLn(SOCKET *sck, string buffer)
{
    return sendToSocket(sck, buffer + "\r\n");
}


//
// get the local ip of a connected socket
//
bool getLocalSockIp(SOCKET *sck, char **peername)
{
    // retrive remote address
    struct sockaddr_in addrin;
    int addrinlen = sizeof( addrin );

    if (getsockname(*sck, (struct sockaddr *) &addrin, (int *)&addrinlen) < 0)
        return false;

    *peername = inet_ntoa(addrin.sin_addr);
    return true;
}

bool getLocalSockIp(SOCKET *sck, string &peername)
{
    char *buf = 0;
    if (!getLocalSockIp(sck, &buf))
        return false;

    peername = "";
    int i = 0;
    while (*(buf + i) != 0)
        peername += *(buf + i++);

    return true;
}


//
// get the remote ip of a connected socket
//
bool getRemoteSockIp(SOCKET *sck, char **peername)
{
    // retrive remote address
    struct sockaddr_in addrin;
    int addrinlen = sizeof( addrin );

    if (getpeername(*sck, (struct sockaddr *) &addrin, (int *)&addrinlen) < 0)
        return false;

    *peername = inet_ntoa(addrin.sin_addr);
    return true;
}
